;;; pgg-gpg.el --- GnuPG support for PGG.

;; Copyright (C) 1999, 2000, 2002, 2003, 2004,
;;   2005, 2006 Free Software Foundation, Inc.

;; Author: Daiki Ueno <ueno@unixuser.org>
;; Symmetric encryption support added by: Sascha Wilde <wilde@sha-bang.de>
;; Created: 1999/10/28
;; Keywords: PGP, OpenPGP, GnuPG

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
;; Boston, MA 02110-1301, USA.

;;; Commentary:

;; NOTE: This file is a backend of pgg.el.  Please do not directly
;; call functions defined in this file from your program.

;;; Code:

(eval-when-compile
  (require 'pgg))

(defgroup pgg-gpg ()
  "GnuPG interface."
  :group 'pgg)

(defcustom pgg-gpg-program "gpg"
  "The GnuPG executable."
  :group 'pgg-gpg
  :type 'string)

(defcustom pgg-gpg-extra-args nil
  "Extra arguments for every GnuPG invocation."
  :group 'pgg-gpg
  :type '(repeat (string :tag "Argument")))

(defcustom pgg-gpg-use-agent nil
  "Whether to use gnupg agent for key caching."
  :group 'pgg-gpg
  :type 'boolean)

(defvar pgg-gpg-user-id nil
  "GnuPG ID of your default identity.")

(defvar pgg-gpg-user-id-alist nil
  "An alist mapping from key ID to user ID.")

(defvar pgg-gpg-read-point nil)
(defvar pgg-gpg-pending-status-list nil)
(defvar pgg-gpg-key-id nil)
(defvar pgg-gpg-passphrase nil)
(defvar pgg-gpg-debug nil)

(defvar pgg-gpg-colons-pub-spec
  '((trust "[^:]")
    (length "[0-9]+" 0 string-to-number)
    (algorithm "[0-9]+" 0 string-to-number)
    (key-id "[^:]+")
    (creation-date "[0-9]+")
    (expiration-date "[0-9]+")
    nil
    (ownertrust "[^:]")
    nil
    nil
    (capability "[escaESCA]*")))

(defvar pgg-gpg-colons-sec-spec
  '((trust "[^:]")
    (length "[0-9]+" 0 string-to-number)
    (algorithm "[0-9]+" 0 string-to-number)
    (key-id "[^:]+")
    (creation-date "[0-9]+")
    (expiration-date "[0-9]+")
    nil
    (ownertrust "[^:]")))

(defvar pgg-gpg-colons-uid-spec
  '((trust "[^:]")
    nil
    nil
    nil
    (creation-date "[0-9]+")
    (expiration-date "[0-9]+")
    (hash "[^:]+")
    nil
    (user-id "[^:]+")))
    
(defun pgg-gpg-start-process (args)
  "This function is for internal use only.

Start pgg-gpg-program in a subprocess with given ARGS.
Return the process object for it."
  (let* ((args
	  (append (list "--no-tty"
			"--status-fd" "1"
			"--command-fd" "0"
			"--yes" ; overwrite
			)
		  (if pgg-gpg-use-agent '("--use-agent"))
		  pgg-gpg-extra-args
		  args))
	 (coding-system-for-write 'binary)
	 (process-connection-type nil)
	 (orig-mode (default-file-modes))
	 (buffer (generate-new-buffer " *pgg-gpg*"))
	 process)
    (with-current-buffer buffer
      (make-local-variable 'pgg-gpg-read-point)
      (setq pgg-gpg-read-point (point-min))
      (make-local-variable 'pgg-gpg-pending-status-list)
      (setq pgg-gpg-pending-status-list nil)
      (make-local-variable 'pgg-gpg-key-id)
      (setq pgg-gpg-key-id nil)
      (make-local-variable 'pgg-gpg-passphrase)
      (setq pgg-gpg-passphrase nil))
    (unwind-protect
	(progn
	  (set-default-file-modes 448)
	  (setq process
		(apply #'start-process "pgg-gpg" buffer pgg-gpg-program args)))
      (set-default-file-modes orig-mode))
    (set-process-filter process #'pgg-gpg-process-filter)
    (set-process-sentinel process #'pgg-gpg-process-sentinel)
    process))

(defun pgg-gpg-process-filter (process input)
  (if pgg-gpg-debug
      (save-excursion
	(set-buffer (get-buffer-create  " *pgg-gpg-debug*"))
	(goto-char (point-max))
	(insert input)))
  (if (buffer-live-p (process-buffer process))
      (save-excursion
	(set-buffer (process-buffer process))
	(goto-char (point-max))
	(insert input)
	(goto-char pgg-gpg-read-point)
	(beginning-of-line)
	(while (looking-at ".*\n")	;the input line is finished
	  (save-excursion
	    (if (looking-at "\\[GNUPG:] \\([A-Z_]+\\)\\>.*")
		(let* ((status (match-string 1))
		       (symbol (intern-soft (concat "pgg-gpg-status-"
						    status))))
		  (if (member status pgg-gpg-pending-status-list)
		      (setq pgg-gpg-pending-status-list nil))
		  (if (and symbol
			   (fboundp symbol))
		      (funcall symbol process (buffer-substring
					       (match-beginning 1)
					       (match-end 0)))))))
	  (forward-line))
	(setq pgg-gpg-read-point (point)))))

(defun pgg-gpg-process-sentinel (process status)
  (if (buffer-live-p (process-buffer process))
      (save-excursion
	(set-buffer (process-buffer process))
	(when pgg-gpg-passphrase
	  (fillarray pgg-gpg-passphrase 0)
	  (setq pgg-gpg-passphrase nil))
	;; Copy the contents of process-buffer to pgg-errors-buffer.
	(set-buffer (get-buffer-create pgg-errors-buffer))
	(buffer-disable-undo)
	(erase-buffer)
	(insert-buffer-substring (process-buffer process)))))

(defun pgg-gpg-read-output-file (output-file-name)
  (set-buffer (get-buffer-create pgg-output-buffer))
  (buffer-disable-undo)
  (erase-buffer)
  (if (file-exists-p output-file-name)
      (let ((coding-system-for-read (if pgg-text-mode
					'raw-text
				      'binary)))
	(insert-file-contents output-file-name))))

(defun pgg-gpg-wait-for-status (process status-list)
  (with-current-buffer (process-buffer process)
    (setq pgg-gpg-pending-status-list status-list)
    (while (and (eq (process-status process) 'run)
		pgg-gpg-pending-status-list)
      (accept-process-output process 1))))

(defun pgg-gpg-wait-for-completion (process)
  (process-send-eof process)
  (while (eq (process-status process) 'run)
    ;; We can't use accept-process-output instead of sit-for here
    ;; because it may cause an interrupt during the sentinel execution.
    (sit-for 0.1)))

(defun pgg-gpg-status-USERID_HINT (process line)
  (if (string-match "\\`USERID_HINT \\([^ ]+\\) \\(.*\\)" line)
      (let* ((key-id (match-string 1 line))
	     (user-id (match-string 2 line))
	     (entry (assoc key-id pgg-gpg-user-id-alist)))
	(if entry
	    (setcdr entry user-id)
	  (setq pgg-gpg-user-id-alist (cons (cons key-id user-id)
					    pgg-gpg-user-id-alist))))))

(defun pgg-gpg-status-NEED_PASSPHRASE (process line)
  (if (string-match "\\`NEED_PASSPHRASE \\([^ ]+\\)" line)
      (setq pgg-gpg-key-id (match-string 1 line))))

(defun pgg-gpg-status-NEED_PASSPHRASE_SYM (process line)
  (setq pgg-gpg-key-id 'SYM))

(defun pgg-gpg-status-NEED_PASSPHRASE_PIN (process line)
  (setq pgg-gpg-key-id 'PIN))

(defun pgg-gpg-status-GET_HIDDEN (process line)
  (let ((entry (assoc pgg-gpg-key-id pgg-gpg-user-id-alist)))
    (if (setq pgg-gpg-passphrase
	      (if (eq pgg-gpg-key-id 'SYM)
		  (pgg-read-passphrase
		   "GnuPG passphrase for symmetric encryption: ")
		(pgg-read-passphrase
		 (format "GnuPG passphrase for %s: "
			 (if entry
			     (cdr entry)
			   pgg-gpg-key-id))
		 (if (eq pgg-gpg-key-id 'PIN)
		     "PIN"
		   pgg-gpg-key-id))))
	(process-send-string process (concat pgg-gpg-passphrase "\n")))))

(defun pgg-gpg-status-GOOD_PASSPHRASE (process line)
  (when (and pgg-gpg-passphrase
	     (stringp pgg-gpg-key-id))
    (pgg-add-passphrase-to-cache pgg-gpg-key-id pgg-gpg-passphrase)
    (setq pgg-gpg-passphrase nil)))

(defun pgg-gpg-status-BAD_PASSPHRASE (process line)
  (when pgg-gpg-passphrase
    (fillarray pgg-gpg-passphrase 0)
    (setq pgg-gpg-passphrase nil)))

(defun pgg-gpg-lookup-key (string &optional type)
  "This function is for internal use only.

Search keys associated with STRING."
  (let ((args (list "--with-colons" "--no-greeting" "--batch"
		    (if type "--list-secret-keys" "--list-keys")
		    string)))
    (with-temp-buffer
      (apply #'call-process pgg-gpg-program nil t nil args)
      (goto-char (point-min))
      (if (re-search-forward "^\\(sec\\|pub\\):[^:]*:[^:]*:[^:]*:\\([^:]*\\)"
			     nil t)
	  (substring (match-string 2) 8)))))

(defun pgg-gpg-list-keys (name &optional secret)
  "This function is for internal use only.

Search keys associated with STRING."
  (let ((args (list "--with-colons" "--no-greeting" "--batch"
		    "--fixed-list-mode"
		    (if secret "--list-secret-keys" "--list-keys")
		    name))
	keys
	type
	symbol
	pointer)
    (with-temp-buffer
      (apply #'call-process pgg-gpg-program nil (list t nil) nil args)
      (goto-char (point-min))
      (while (looking-at "\\([a-z][a-z][a-z]\\):\\(.*\\)")
	(setq type (match-string 1)
	      symbol (intern-soft (format "pgg-gpg-colons-%s-spec" type)))
	(if (member type '("pub" "sec"))
	    (setq keys (cons nil keys)))
	(if (and symbol
		 (boundp symbol))
	    (setcar keys (cons (cons (intern type)
				     (pgg-gpg-parse-colons
				      (symbol-value symbol)
				      (match-string 2)))
			       (car keys))))
	(forward-line)))
    (setq pointer keys)
    (while pointer
      (setcar pointer (nreverse (car pointer)))
      (setq pointer (cdr pointer)))
    (nreverse keys)))

(defun pgg-gpg-parse-colons (alist string)
  (let ((index 0)
	result)
    (while (and alist
		(or (null (car alist))
		    (eq index
			(string-match
			 (concat "\\(" (nth 1 (car alist)) "\\)?:")
			 string index))))
      (if (car alist)
	  (progn
	    (setq index (match-end 0))
	    (if (match-beginning 1)
		(setq result (cons (cons (car (car alist))
					 (funcall
					  (or (nth 3 (car alist)) #'identity)
					  (match-string
					   (1+ (or (nth 2 (car alist)) 0))
					   string)))
				   result))))
	(setq index (1+ index)))
      (setq alist (cdr alist)))
    (nreverse result)))

(defun pgg-gpg-encrypt-region (start end recipients &optional sign passphrase)
  "This function is for internal use only.

Encrypt the current region between START and END.

If optional argument SIGN is non-nil, do a combined sign and encrypt.

Obsolete feature: if optional PASSPHRASE is not specified, it will be
obtained from the passphrase cache or user."
  (let* ((pgg-gpg-user-id (or pgg-gpg-user-id pgg-default-user-id))
	 (output-file-name (pgg-make-temp-file "pgg-output"))
	 (args
	  (append
	   (list "--armor" "--always-trust" "--encrypt"
		 "--output" output-file-name)
	   (if passphrase '("--passphrase-fd" "0"))
	   (if pgg-text-mode '("--textmode"))
	   (if sign (list "--sign" "--local-user" pgg-gpg-user-id))
	   (if recipients
	       (apply #'nconc
		      (mapcar (lambda (rcpt)
				(list "-r" rcpt))
			      (if pgg-encrypt-for-me
				  (cons pgg-gpg-user-id recipients)
				recipients))))))
	 process status)
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process args))
	  (if (and passphrase
		   (eq (process-status process) 'run))
	      (process-send-string process (concat passphrase "\n")))
	  (if (and sign (not pgg-gpg-use-agent))
	      (pgg-gpg-wait-for-status process
				       '("BEGIN_SIGNING" "GOOD_PASSPHRASE")))
	  (if (eq (process-status process) 'run)
	      (process-send-region process start end))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer pgg-errors-buffer))
	    (goto-char (point-max))
	    (setq status (not (null (re-search-backward
				     "^\\[GNUPG:] END_ENCRYPTION\\>"
				     nil t)))))
	  (if status
	      (pgg-gpg-read-output-file output-file-name))
	  status)
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-encrypt-symmetric-region (start end &optional passphrase)
  "This function is for internal use only.

Encrypt the current region between START and END with symmetric cipher.

If optional PASSPHRASE is not specified, it will be
obtained from the passphrase cache or user."
  (let* ((output-file-name (pgg-make-temp-file "pgg-output"))
	 (args
	  (append (list "--armor" "--symmetric"
			"--output" output-file-name)
		  (if passphrase '("--passphrase-fd" "0"))
		  (if pgg-text-mode '("--textmode"))))
	 process status)
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process args))
	  (if (and passphrase
		   (eq (process-status process) 'run))
	      (process-send-string process (concat passphrase "\n")))
	  (pgg-gpg-wait-for-status process '("BEGIN_ENCRYPTION"))
	  (if (eq (process-status process) 'run)
	      (process-send-region process start end))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer pgg-errors-buffer))
	    (goto-char (point-max))
	    (setq status (not (null (re-search-backward
				     "^\\[GNUPG:] END_ENCRYPTION\\>"
				     nil t)))))
	  (if status
	      (pgg-gpg-read-output-file output-file-name))
	  status)
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-decrypt-region (start end &optional passphrase)
  "This function is for internal use only.

Decrypt the current region between START and END.

Obsolete feature: if optional PASSPHRASE is not specified, it will be
obtained from the passphrase cache or user."
  (let* ((input-file-name (pgg-make-temp-file "pgg-input"))
	 (output-file-name (pgg-make-temp-file "pgg-output"))
	 (args (append
		(list "--decrypt" "--output" output-file-name input-file-name)
		(if passphrase '("--passphrase-fd" "0"))))
	 process status)
    (unwind-protect
	(progn
	  (write-region start end input-file-name)
	  (setq process (pgg-gpg-start-process args))
	  (if (and passphrase
		   (eq (process-status process) 'run))
	      (process-send-string process (concat passphrase "\n")))
	  (pgg-gpg-wait-for-status process '("BEGIN_DECRYPTION"))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer-create pgg-errors-buffer))
	    (goto-char (point-max))
	    (setq status (not (null (re-search-backward
				     "^\\[GNUPG:] DECRYPTION_OKAY\\>"
				     nil t)))))
	  (if status
	      (pgg-gpg-read-output-file output-file-name))
	  status)
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p input-file-name)
	  (delete-file input-file-name))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-sign-region (start end &optional cleartext passphrase)
  "This function is for internal use only.

Make detached signature from text between START and END.

Obsolete feature: if optional PASSPHRASE is not specified, it will be
obtained from the passphrase cache or user."
  (let* ((output-file-name (pgg-make-temp-file "pgg-output"))
	 (pgg-gpg-user-id (or pgg-gpg-user-id pgg-default-user-id))
	 (args
	  (append (list (if cleartext "--clearsign" "--detach-sign")
			"--armor" "--verbose"
			"--local-user" pgg-gpg-user-id
			"--output" output-file-name)
		  (if passphrase '("--passphrase-fd" "0"))
		  (if pgg-text-mode '("--textmode"))))
	 process status)
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process args))
	  (if (and passphrase
		   (eq (process-status process) 'run))
	      (process-send-string process (concat passphrase "\n")))
	  (unless pgg-gpg-use-agent
	    (pgg-gpg-wait-for-status process
				     '("BEGIN_SIGNING" "GOOD_PASSPHRASE")))
	  (if (eq (process-status process) 'run)
	      (process-send-region process start end))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer-create pgg-errors-buffer))
	    (goto-char (point-max))
	    (setq status (not (null (re-search-backward
				     "^\\[GNUPG:] SIG_CREATED\\>"
				     nil t)))))
	  (if status
	      (pgg-gpg-read-output-file output-file-name))
	  status)
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-verify-region (start end &optional signature)
  "This function is for internal use only.

Verify region between START and END as the detached signature SIGNATURE."
  (let* ((output-file-name (pgg-make-temp-file "pgg-output"))
	 (args (list "--verify" "--output" output-file-name))
	 process status)
    (if signature
	(setq args (append args (list signature))))
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process (append args '("-"))))
	  (if (eq (process-status process) 'run)
	      (process-send-region process start end))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer-create pgg-errors-buffer))
	    (goto-char (point-max))
	    (setq status (not (null (re-search-backward
				     "^\\[GNUPG:] GOODSIG\\>"
				     nil t)))))
	  (if status
	      (pgg-gpg-read-output-file output-file-name))
	  status)
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-insert-key ()
  "This function is for internal use only.

Insert public key at point."
  (let* ((pgg-gpg-user-id (or pgg-gpg-user-id pgg-default-user-id))
	 (output-file-name (pgg-make-temp-file "pgg-output"))
	 (args (list "--export" "--armor" "--output" output-file-name
		     pgg-gpg-user-id))
	 process status)
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process args))
	  (pgg-gpg-wait-for-completion process)
	  (pgg-gpg-read-output-file output-file-name)
	  (insert-buffer-substring pgg-output-buffer))
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process)))
      (if (file-exists-p output-file-name)
	  (delete-file output-file-name)))))

(defun pgg-gpg-snarf-keys-region (start end)
  "This function is for internal use only.

Add all public keys in region between START and END to the keyring."
  (let ((args '("--import" "-"))
	process)
    (unwind-protect
	(progn
	  (setq process (pgg-gpg-start-process args))
	  (if (eq (process-status process) 'run)
	      (process-send-region process start end))
	  (pgg-gpg-wait-for-completion process)
	  (save-excursion
	    (set-buffer (get-buffer-create pgg-errors-buffer))
	    (goto-char (point-max))
	    (not (null (re-search-backward
			"^\\[GNUPG:] IMPORT_RES\\>"
			nil t)))))
      (if (and process (buffer-live-p (process-buffer process)))
	  (kill-buffer (process-buffer process))))))

(provide 'pgg-gpg)

;;; arch-tag: 2aa5d5d8-93a0-4865-9312-33e29830e000
;;; pgg-gpg.el ends here
